/******************************************************************************

  Copyright (c) 2007,2008 Turku PET Centre

  Library:      imgdecay
  Description:  Functions for working with physical decay and isotopes in IMG.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details:
  http://www.gnu.org/copyleft/lesser.html

  You should have received a copy of the GNU Lesser General Public License
  along with this library/program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 

  Turku PET Centre, Turku, Finland, http://www.turkupetcentre.fi/

  Modification history:
  2007-02-01 Vesa Oikonen
    First created this file in libtpcimgio 1.2.0.
    Functions now apply libtpcmisc 1.2.0 as far as possible.
    Added function imgSetDecayCorrFactors().
  2008-07-07 VO
    Decay correction functions check (superficially) that image data contains
    frames times.


******************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <ctype.h>
/*****************************************************************************/
#include "halflife.h"
/*****************************************************************************/
#include "include/img.h"
#include "include/imgdecay.h"
/*****************************************************************************/

/*****************************************************************************/
/*!
 *  Corrects (mode=1) or removes correction (mode=0) for physical decay.
 *  Removal is based on existing decay correction factors, when possible.
 *
 * @param image pointer to IMG data
 * @param mode 0=Remove decay correction; 1=Correct for decay
 * @return 0 if ok, 1 image status is not 'occupied', 2 decay already corrected/not corrected,
 * 3 image frame times missing
 */
int imgDecayCorrection(IMG *image, int mode) {
  int pi, fi, i, j;
  float lambda;
  float cf, dur;

  /* Check for arguments */
  if(image->status!=IMG_STATUS_OCCUPIED) return(1);
  if(image->isotopeHalflife<=0.0) return(1);
  /* Existing/nonexisting decay correction is an error */
  if(mode==1 && image->decayCorrected!=0) return(2);
  if(mode==0 && image->decayCorrected==0) return(2);

  /* All time frames */
  for(fi=0; fi<image->dimt; fi++) {
    dur=image->end[fi]-image->start[fi];
    if(image->end[fi]>0.0) {
      if(mode==0 && image->decayCorrFactor[fi]>1.000001) {
        /* if decay correction is to be removed, and factor is known, then use it */
        cf=1.0/image->decayCorrFactor[fi];
      } else {
        lambda=hl2lambda(image->isotopeHalflife); if(lambda<0.0) return(1);
        if(mode==0) lambda=-lambda; /* remove decay correction by giving negative lambda */
        if(fi==image->dimt-1 && image->end[fi]<=0.0) return(3);
        cf=hlLambda2factor_float(lambda, image->start[fi], dur);
      }
      if(IMG_TEST) printf("applied_dc_factor[%d] := %g\n", fi+1, cf);
      /* Set decay correction factor inside IMG for future */
      if(mode==0) {
        image->decayCorrFactor[fi]=1.0;
      } else {
        image->decayCorrFactor[fi]=cf;
      }
      /* All planes, all matrices */
      for(pi=0; pi<image->dimz; pi++)
        for(i=0; i<image->dimy; i++)
          for(j=0; j<image->dimx; j++)
            image->m[pi][i][j][fi]*=cf;
      image->decayCorrected=mode; /* in some cases left unchanged! */
    }
  } /* next frame */
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/*!
 * Returns pointer to string describing the isotope in image data
 *
 * @param img image stucture
 * @return pointer to string
 */
char *imgIsotope(IMG *img) {
  return(hlIsotopeCode(hlIsotopeFromHalflife(img->isotopeHalflife/60.0)));
}
/*****************************************************************************/

/*****************************************************************************/
/*!
 * Sets (mode=1) or removes (mode=0) decay correction factors in IMG.
 * IMG pixel data is not changed.
 *
 * @param image pointer to IMG data
 * @param mode factors are calculated for decay correction (1) or for removing decay
 * correction (0)
 * @return 0 if ok, 1 image status is not 'occupied', 2 invalid exponent value,
 * 3 image frame times are missing
 */
int imgSetDecayCorrFactors(IMG *image, int mode) {
  int fi;
  float lambda, cf, dur;

  /* Check for arguments */
  if(image->status!=IMG_STATUS_OCCUPIED) return(1);
  if(image->isotopeHalflife<=0.0) return(1);

  /* Check that image contains frame times */
  if(mode!=0 && image->end[image->dimt-1]<=0.0) return(3);

  /* All time frames */
  for(fi=0; fi<image->dimt; fi++) {
    if(mode==0) {
      image->decayCorrFactor[fi]=1.0;
    } else {
      dur=image->end[fi]-image->start[fi];
      if(image->end[fi]>0.0) {
        lambda=hl2lambda(image->isotopeHalflife); if(lambda<0.0) return(2);
        cf=hlLambda2factor_float(lambda, image->start[fi], dur);
        image->decayCorrFactor[fi]=cf;
      }
    }
  } /* next frame */
  image->decayCorrected=mode;
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/

